package at.tugraz.iaik.p2p.commons

//import iaik.security.ssl.ECCelerateProvider
//import iaik.security.ssl.SecurityProvider
import iaik.asn1.structures.AlgorithmID
import iaik.security.ec.common.*
import iaik.security.ec.errorhandling.DecodingException
import iaik.security.ec.errorhandling.InvalidCurveParameterSpecException
import iaik.security.ec.provider.ECCelerate
import iaik.security.provider.IAIK
import iaik.x509.X509Certificate
import iaik.x509.X509ExtensionException
import iaik.x509.extensions.BasicConstraints
import iaik.x509.extensions.CRLDistributionPoints
import iaik.x509.extensions.ExtendedKeyUsage
import iaik.x509.extensions.KeyUsage
import iaik.x509.extensions.netscape.NetscapeCertType
import java.io.IOException
import java.math.BigInteger
import java.nio.ByteBuffer
import java.security.*
import java.security.interfaces.ECPrivateKey
import java.security.interfaces.ECPublicKey
import java.util.*
import javax.crypto.Cipher
import kotlin.experimental.and
import kotlin.experimental.or

/**
 * Created by Bernd Prünster <bernd.pruenster@iaik.tugraz.at>
 */

fun Signature.sign(data: ByteBuffer): ByteArray {
    update(data)
    return sign()
}

fun Signature.verify(data: ByteBuffer, sigBytes: ByteArray) {
    update(data)
    if (!verify(sigBytes)) {
        throw SignatureException("Invalid signature")
    }
}

fun ECPublicKey.encode() = CommonCrypto.encode(this)

object CommonCrypto {

    @JvmStatic
    private val sigAlg = "SHA3-256withECDSA"
    private val sigAlgED = "EdDSA25519"

    @JvmStatic
    val TLS_CIPHERS: Array<String>
        get() = getCipherSuites("EC") + getCipherSuites("RSA")
    val RND: SecureRandom = SecureRandom.getInstance("SHA512PRNG-SP80090", IAIK.getInstance())

    init {
        System.setProperty("java.security.egd","file:/dev/urandom")
        IAIK.addAsProvider(true)
        ECCelerate.addAsProvider(true)
    }


    @Throws(CryptoException::class)
    @JvmStatic
    fun genKP(algorithm: String, len: Int): KeyPair {
        val ec: KeyPairGenerator
        try {
            ec = KeyPairGenerator.getInstance(
                algorithm,
                if (algorithm == "EC") ECCelerate.getInstance() else IAIK.getInstance()
            )
            if (algorithm == "EC") ec.initialize(ECStandardizedParameterFactory.getParametersByBitLength(len), RND)
            else ec.initialize(len)
            return ec.generateKeyPair()
        } catch (e: InvalidAlgorithmParameterException) {
            throw CryptoException(e)
        } catch (e: NoSuchAlgorithmException) {
            throw CryptoException(e)
        }

    }

    @Throws(CryptoException::class)
    @JvmStatic
    fun genKP(ed: Boolean = true): ECKeyPair {
        val ec: KeyPairGenerator
        try {
            ec = if (ed) {
                val kpg = KeyPairGenerator.getInstance(sigAlgED, ECCelerate.getInstance())
                kpg.initialize(EdStandardizedParameterFactory.getParametersByName(sigAlgED), RND)
                kpg
            } else {
                val kpg = KeyPairGenerator.getInstance("EC", ECCelerate.getInstance())
                kpg.initialize(ECStandardizedParameterFactory.getParametersByBitLength(256), RND)
                kpg
            }
            return ECKeyPair(ec.generateKeyPair())
        } catch (e: InvalidAlgorithmParameterException) {
            throw CryptoException(e)
        } catch (e: NoSuchAlgorithmException) {
            throw CryptoException(e)
        }

    }

    @Throws(CryptoException::class)
    @JvmStatic
    fun genCert(
        pub: PublicKey,
        subject: Principal,
        issuerCN: Principal = subject,
        signingKey: PrivateKey,
        ca: Boolean,
        from: Date,
        to: Date,
        crlDistributionPoints: Array<String>? = null
    ): java.security.cert.X509Certificate {

        val cert = X509Certificate()
        cert.serialNumber = BigInteger.valueOf(RND.nextLong())
        cert.subjectDN = subject
        try {
            cert.publicKey = pub

        } catch (e: InvalidKeyException) {
            throw CryptoException(e)
        }

        cert.issuerDN = issuerCN
        cert.setValidNotBefore(from)
        cert.setValidNotAfter(to)
        // CRL distribution point
        crlDistributionPoints?.let {
            if (!crlDistributionPoints.isEmpty()) {
                val point = iaik.asn1.structures.DistributionPoint(crlDistributionPoints)
                point.reasonFlags = iaik.asn1.structures.DistributionPoint.keyCompromise
                cert.addExtension(CRLDistributionPoints(point))
            }
        }

        try {
            if (ca) {
                val basicConstraints = BasicConstraints(true, 1)
                basicConstraints.isCritical = true
                cert.addExtension(basicConstraints)
                cert.addExtension(NetscapeCertType(NetscapeCertType.SSL_CA or NetscapeCertType.S_MIME_CA))

                val keyUsage = KeyUsage(KeyUsage.digitalSignature or KeyUsage.cRLSign or KeyUsage.keyCertSign)
                keyUsage.isCritical = true
                cert.addExtension(keyUsage)
            } else {
                cert.addExtension(KeyUsage(KeyUsage.digitalSignature or KeyUsage.keyEncipherment))
                cert.addExtension(ExtendedKeyUsage(arrayOf(ExtendedKeyUsage.serverAuth, ExtendedKeyUsage.clientAuth)))
            }
        } catch (e: X509ExtensionException) {
            throw CryptoException(e)
        }

        try {
            cert.sign(AlgorithmID.getAlgorithmID(signingKey.getSignatureAlgorithm()), signingKey)
        } catch (e: Exception) {
            throw CryptoException(e)
        }

        return cert
    }



    @Throws(CryptoException::class)
    @JvmStatic
    fun encode(pubKey: ECPublicKey): ByteArray {
        val params: ECParameterSpec
        if (pubKey is EdPublicKey) {
            params = pubKey.params
        } else {
            try {
                params = ECParameterSpec.getParameterSpec(pubKey.params)
            } catch (e: InvalidCurveParameterSpecException) {
                throw CryptoException(e)
            }
        }

        var compressedPoint = params.curve.encodePoint(pubKey.w, PointEncoders.COMPRESSED)
        if (pubKey is EdPublicKey)
            return compressedPoint
        compressedPoint[0] =
                ((compressedPoint[0] and 1.toByte()) or 2.toByte())

        return compressedPoint
    }

    @Throws(CryptoException::class)
    @JvmStatic
    fun decode(encoded: ByteArray): ECPublicKey {
        val params: ECParameterSpec
        try {
            params = specFromPubKey(encoded) as ECParameterSpec
        } catch (e: InvalidCurveParameterSpecException) {
            throw CryptoException(e)
        }

        if (params is EdParameterSpec) {
            return EdPublicKey(EdPublicKeySpec(encoded, params))
        } else
            encoded[0] = (encoded[0] or 2.toByte())
        try {
            val ecPoint = params.curve.decodePoint(
                if (params is EdParameterSpec) encoded.copyOfRange(
                    1,
                    encoded.size
                ) else encoded
            )
            return iaik.security.ec.common.ECPublicKey(params, ecPoint)
        } catch (e: DecodingException) {
            throw CryptoException(e)
        }

    }

    @Throws(CryptoException::class)
    private fun specFromPubKey(encodedKey: ByteArray): java.security.spec.ECParameterSpec {
        if (encodedKey.size == 32)
            return EdParameterSpec.getParameterSpec(EdStandardizedParameterFactory.getParametersByName("EdDSA25519"))
        val typ = (encodedKey[0] and 254.toByte())
        return when (typ) {
            2.toByte() -> ECParameterSpec.getParameterSpec(ECStandardizedParameterFactory.getParametersByBitLength(256))
            else -> throw CryptoException("Unknown Key Type: $typ")
        }
    }


    @Throws(CryptoException::class)
    @JvmStatic
    fun encrypt(key: ECPublicKey, input: ByteArray): ByteArray {
        val cipher: Cipher
        try {
            cipher = Cipher.getInstance("ECIES")
            cipher.init(Cipher.ENCRYPT_MODE, key)
            return cipher.doFinal(input)
        } catch (e: Exception) {
            throw CryptoException(e)
        }

    }


    @Throws(CryptoException::class)
    @JvmStatic
    fun createSignature(key: ECPrivateKey): Signature {
        val signature = Signature.getInstance(if (key is EdPrivateKey) sigAlgED else sigAlg)
        signature.initSign(key)
        return signature
    }


    @Throws(CryptoException::class)
    @JvmStatic
    fun sign(key: ECPrivateKey, data: ByteArray): ByteArray {
        try {
            val signature = Signature.getInstance(sigAlg)
            signature.initSign(key)
            signature.update(data)
            return signature.sign()
        } catch (e: Exception) {
            throw CryptoException(e)
        }

    }

    @JvmStatic
    @Throws(CryptoException::class)
    fun createVerify(key: ECPublicKey): Signature {
        val signature = Signature.getInstance(if (key is EdPublicKey) sigAlgED else sigAlg)
        signature.initVerify(key)
        return signature
    }


    @Throws(CryptoException::class)
    @JvmStatic
    fun validate(data: ByteArray, sigBytes: ByteArray, key: ECPublicKey) {
        try {
            val signature = Signature.getInstance(sigAlg)
            signature.initVerify(key)
            signature.update(data)
            if (!signature.verify(sigBytes)) {
                throw SignatureException("Invalid signature")
            }
        } catch (e: InvalidKeyException) {
            throw CryptoException(e)
        } catch (e: NoSuchAlgorithmException) {
            throw CryptoException(e)
        } catch (e: IOException) {
            throw CryptoException(e)
        } catch (e: SignatureException) {
            throw CryptoException(e)
        }

    }


    @Throws(CryptoException::class)
    @JvmStatic
    fun sha3(data: ByteBuffer) = MessageDigest.getInstance("SHA3-256").apply { update(data) }.digest()

    @Throws(CryptoException::class)
    @JvmStatic
    fun sha3(data: ByteArray) = MessageDigest.getInstance("SHA3-256").apply { update(data) }.digest()

    @Throws(CryptoException::class)
    @JvmStatic
    fun sha1(data: ByteArray) = MessageDigest.getInstance("SHA-1").apply { update(data) }.digest()

    @Throws(CryptoException::class)
    @JvmStatic
    fun sha1(data: ByteBuffer) = MessageDigest.getInstance("SHA-1").apply { update(data) }.digest()


    @JvmStatic
    fun init() {
        sha3(byteArrayOf(0))
    }
}

class ECKeyPair internal constructor(kp: KeyPair) {
    /**
     * Returns a reference to the public key component of this key pair.
     *
     * @return a reference to the public key.
     */
    val public: ECPublicKey = kp.public as ECPublicKey
    /**
     * Returns a reference to the private key component of this key pair.
     *
     * @return a reference to the private key.
     */
    val private: ECPrivateKey = kp.private as ECPrivateKey


    fun toKP(): KeyPair {
        return KeyPair(public, private)
    }

}


fun PrivateKey.getSignatureInstance(): Signature {
    return Signature.getInstance(getSignatureAlgorithm())
}

fun PrivateKey.getSignatureAlgorithm(): String {
    if (algorithm == "RSA") return "SHA3-256/RSA"
    if (algorithm == "EC") return "SHA3-256/ECDSA"
    throw CryptoException("Algorithm $algorithm is not supported!")
}

fun PrivateKey.getTLSCipherSuites(): Array<String> {
    return getCipherSuites(algorithm)
}

fun PublicKey.getTLSCipherSuites(): Array<String> {
    return getCipherSuites(algorithm)
}


private fun getCipherSuites(algorithm: String): Array<String> {
    val alg: String
    if (algorithm == "RSA") alg = "RSA"
    else if (algorithm == "EC") alg = "ECDSA"
    else throw CryptoException("Algorithm $algorithm not supported!")

    return arrayOf(
        "TLS_ECDHE_${alg}_WITH_AES_256_CBC_SHA384",
        "TLS_ECDHE_${alg}_WITH_AES_128_GCM_SHA256",
        "TLS_ECDHE_${alg}_WITH_AES_128_CBC_SHA256"
    )


}

data class KeyAndCertChain(val key: PrivateKey, val chain: Array<java.security.cert.X509Certificate>)

class CryptoException(msg: String? = null, e: Throwable? = null) : Throwable(msg, e) {
    constructor(e: Throwable) : this(null, e)
}
