package at.tugraz.iaik.p2p.evt

import at.tugraz.iaik.p2p.commons.CommonCrypto
import at.tugraz.iaik.p2p.evt.msg.*
import at.tugraz.iaik.p2p.hkad.config.Config
import at.tugraz.iaik.p2p.hkad.id.Identifier
import at.tugraz.iaik.p2p.hkad.msg.ExtendedMessage
import at.tugraz.iaik.p2p.hkad.node.LocalNode
import at.tugraz.iaik.p2p.hkad.node.Node
import kotlinx.coroutines.GlobalScope
import kotlinx.coroutines.delay
import kotlinx.coroutines.launch
import kotlinx.coroutines.sync.Mutex
import mu.KotlinLogging
import java.util.*
import java.util.concurrent.ConcurrentLinkedQueue
import java.util.concurrent.atomic.AtomicInteger

enum class Error{
    UNKNOWN_EVT,UNKNOWN_MSG,OUTDATED_STATE,INCONSISTENT_STATE,DETACHED_EVT,NETWORK
}

class EvtNode(private val config: Config) {
    private val node = LocalNode(config, incomingMsgListener = { listen(it) })
    val logger = KotlinLogging.logger("E ${node.id}")
    val identifier = node.id


    private val criterionMap = mutableMapOf<Identifier, MutableMap<ByteArray, Event>>()
    private val criterionHeads = mutableMapOf<Identifier, Event>()
    private val mapLock = Mutex()
    suspend fun bootstrap(node: EvtNode) = this.node.bootstrap(node.node)


    suspend fun pub(evt: Event): Transmission {
        val pubMessage = PubMessage(CommonCrypto.RND.nextLong(), node, evt)
        return transmit(getResponsibleNodes(evt), pubMessage)
    }

    //TODO verify criterion of responses
    suspend fun poll(criterion: Identifier, sinceDigest: ByteArray): Transmission {
        val pollMessage = PollMessage(CommonCrypto.RND.nextLong(), node, criterion, sinceDigest)
        return transmit(getResponsibleNodes(criterion), pollMessage)
    }

    //TODO verify criterion of responses
    suspend fun peek(criterion: Identifier): Transmission {
        val peekMessage = PeekMessage(CommonCrypto.RND.nextLong(), node, criterion)
        return transmit(getResponsibleNodes(criterion), peekMessage)
    }

    private suspend fun listen(msg: ExtendedMessage): ExtendedMessage {
        logger.trace { "received ${msg.javaClass.simpleName}" }

        when (msg) {

            is PubMessage -> {
                val event = msg.evt
                try {
                    mapLock.lock(criterionMap)

                    if (event.prev contentEquals byteArrayOf()) {
                        logger.trace { "pubMessage is a genesis event" }
                        if (criterionMap.containsKey(event.criterion)) {
                            logger.warn { "genesis pubMessage received for existing chain" }
                            return ErrorMessage(
                                msg.seqId,
                                node,
                                Error.DETACHED_EVT,
                                "genesis event sent, although a chain already exists"
                            )
                        }
                    }

                    if (!criterionMap.containsKey(event.criterion)) {
                        criterionMap[event.criterion] = mutableMapOf(Pair(event.digest, event))
                        criterionHeads[event.criterion] = event
                    }

                    val head = criterionHeads[event.criterion]!!
                    val evtMap = criterionMap[event.criterion]!!

                    if (event.prev contentEquals byteArrayOf()) ;
                    //previous event is not current head
                    else if (!head.digest.contentEquals(event.prev)) {
                        if (evtMap.containsKey(event.prev)) {
                            return ErrorMessage(msg.seqId, node, Error.OUTDATED_STATE, "Previous event not current")
                        } else {
                            val (success, messages) = poll(event.criterion, head.digest)
                            if (!success)
                                return ErrorMessage(
                                    msg.seqId,
                                    node,
                                    Error.INCONSISTENT_STATE,
                                    "previous event not known, could not update"
                                )
                            val msg = messages.filter { it is PollResultMessage }
                                .firstOrNull { (it as PollResultMessage).evtChain.firstOrNull { it.digest contentEquals event.prev } != null } as PollResultMessage

                            criterionHeads[event.criterion] = msg.evtChain[0]
                            evtMap.putAll(msg.evtChain.map { Pair(it.digest, it) })
                            if (!msg.evtChain[0].digest.contentEquals(event.prev))
                                return ErrorMessage(msg.seqId, node, Error.OUTDATED_STATE, "Previous event not current")
                        }
                    }
                    criterionHeads[event.criterion] = event
                    evtMap[event.digest] = event
                    logger.trace { "storing event, head now at $event" }
                    return PubAckMessage(msg.seqId, node, event.digest)

                } finally {
                    mapLock.unlock()
                }
            }
            is PeekMessage -> return PeekResultMessage(
                msg.seqId,
                node,
                criterionHeads[msg.criterionHash]?.digest ?: ByteArray(0)
            )
            is PollMessage -> {
                val event = criterionHeads[msg.criterionHash]
                val events = criterionMap[msg.criterionHash]
                if ((event == null) || (events == null) || (event.digest contentEquals msg.sinceDigest)) {
                    return PollResultMessage(msg.seqId, node, arrayOf())
                }
                if (msg.sinceDigest contentEquals byteArrayOf()) {
                    val evtChain = mutableListOf<Event>()
                    var evt = event
                    while (evt != null) {
                        evtChain.add(evt)
                        evt = events[event.prev]
                    }
                    return PollResultMessage(msg.seqId, node, evtChain.toTypedArray())

                } else {
                    val evtChain = mutableListOf<Event>()
                    var evt = event
                    while (!(evt!!.digest contentEquals msg.sinceDigest)) {
                        evtChain.add(evt)
                        evt = events[event.prev]
                        if (evt == null)
                            return ErrorMessage(msg.seqId, node, Error.UNKNOWN_EVT, "'Since' digest unknown")
                    }
                    return PollResultMessage(msg.seqId, node, evtChain.toTypedArray())
                }

            }
        }

        return ErrorMessage(msg.seqId, node, Error.UNKNOWN_MSG, "message not known!")
    }


    private suspend fun getResponsibleNodes(criterion: Identifier): SortedSet<Node> =
        node.get(criterion).closestNodes

    private suspend fun getResponsibleNodes(evt: Event): SortedSet<Node> =
        getResponsibleNodes(evt.criterion)

    data class Transmission(val success: Boolean, val messages: Collection<ExtendedMessage>)

    private suspend fun transmit(
        closest: SortedSet<Node>,
        message: ExtendedMessage
    ): Transmission {
        val succeeded = AtomicInteger()
        val failed = AtomicInteger()
        val q = ConcurrentLinkedQueue<ExtendedMessage>()
        closest.forEach {
            GlobalScope.launch {
                try {
                    val msg = node.actor.send(message, it) as ExtendedMessage
                    q.add(msg)
                    (if (msg !is ErrorMessage) succeeded else failed).incrementAndGet()
                } catch (e: Throwable) {
                    q.add(ErrorMessage(message.seqId, node, Error.NETWORK, e.localizedMessage))
                    failed.incrementAndGet()
                }
            }
        }
        GlobalScope.launch {
            while (!(
                        (succeeded.get() >= Math.ceil(closest.size.toDouble() / 2.0))
                                ||
                                (failed.get() >= Math.floor(closest.size.toDouble() / 2.0))
                        )
            ) {
                delay(5)
            }

        }.join()
        return Transmission(succeeded.get() >= failed.get(), q)
    }

}