package at.tugraz.iaik.p2p.hkad.id

import at.tugraz.iaik.p2p.hkad.ex.IllegalIdBytes
import com.google.common.primitives.UnsignedBytes
import java.nio.ByteBuffer
import java.util.*
import javax.xml.bind.DatatypeConverter
import kotlin.experimental.xor
import kotlin.math.roundToInt


fun ByteBuffer.putIdentifier(id: Identifier) {
    put(id.type.ordinal.toByte())
    val bytes = if (id is ECID) id.publicKeyBytes else id.keyBytes
    putShort(bytes.size.toShort())
    put(bytes)
}

fun ByteBuffer.getIdentifier(): Identifier {
    val type = Identifier.Type.values()[get().toInt()]
    val bytes = ByteArray(getShort().toInt()).apply { this@getIdentifier.get(this) }
    return when (type) {
        Identifier.Type.KAD -> KadID(bytes)
        Identifier.Type.SHA3 -> SHA3ID(bytes)
        Identifier.Type.EC -> ECID(bytes)
        Identifier.Type.U -> UnknownID
    }
}

class KeyBits(val size: Int) : BitSet(size) {
    private val bArr = ByteArray(Math.ceil(size.toDouble() / 8.0).roundToInt())

    constructor(bytes: ByteArray) : this(bytes.size * 8) {
        val bitSet = BitSet.valueOf(bytes)
        for (i in 0 until bitSet.length())
            this[i] = bitSet[i]
    }

    val bytes: ByteArray
        get() {
            val byteArray = toByteArray()
            for (i in 0 until byteArray.size)
                bArr[i] = byteArray[i]
            for (i in byteArray.size until bArr.size)
                bArr[i] = 0
            return bArr
        }

    fun copy(): KeyBits {
        val copy = KeyBits(size)
        for (i in 0 until length())
            copy[i] = this[i]
        return copy
    }
}


abstract class Identifier protected constructor(
    val keyBytes: ByteArray,
    val type: Type
) : Comparable<Identifier> {

    init {
        if(keyBytes.size*8!=type.numBits.toInt())
            throw IllegalIdBytes(keyBytes.size*8, type.numBits.toInt())
    }

    //protected constructor(keyBits: KeyBits, type: Type) : this(keyBits, keyBits.bytes, type)

    enum class Type(val numBits: Short) {
        KAD(160), SHA3(256), EC(256), U(0)
    }




    override fun compareTo(other: Identifier) = cmp.compare(keyBytes, other.keyBytes)

    override fun hashCode() = Arrays.hashCode(keyBytes)


    override fun equals(other: Any?): Boolean {
        if (other == null)
            return false
        if (other !is Identifier) return false
        if (other.type != type)
            return false
        return keyBytes contentEquals other.keyBytes
    }


    val numBits by lazy { type.numBits }

    open infix fun xor(other: Identifier) = ByteArray(keyBytes.size) { keyBytes[it] xor other.keyBytes[it] }


    abstract fun fromData(data: ByteBuffer): Identifier

    open fun distanceTo(other: Identifier): Int {
        val bits = this xor other
        var dist = 0
        val zero: Byte = 0
        for (i in bits.size - 1 downTo 0) {
            if (bits[i] != zero)
                break
            dist += 8
        }
        val byte = bits[dist / 8].toInt()
        for (i in 7 downTo 0) {
            dist++
            if (byte.shr(i) and 1 == 1) break
        }
        return type.numBits-dist
    }

    abstract fun generateBydistance(dist: Int): Identifier

    protected fun bitsByDistance(dist: Int): ByteArray {
        val byteIndex = dist / 8
        val bitIndex = dist % 8

        val distBitSet = keyBytes.copyOf()
        distBitSet[byteIndex] = distBitSet[byteIndex] xor (1 shl bitIndex).toByte()

        return distBitSet
    }

    override fun toString() = DatatypeConverter.printHexBinary(keyBytes)

    companion object {
        private val cmp = UnsignedBytes.lexicographicalComparator()

    }
}

object UnknownID : Identifier(ByteArray(0), Type.U) {
    override fun fromData(data: ByteBuffer) = this
    override fun generateBydistance(dist: Int) = this
}