package at.tugraz.iaik.p2p.hkad.msg

import at.tugraz.iaik.p2p.hkad.id.getIdentifier
import at.tugraz.iaik.p2p.hkad.id.putIdentifier
import at.tugraz.iaik.p2p.hkad.node.Node
import at.tugraz.iaik.p2p.hkad.node.getNode
import at.tugraz.iaik.p2p.hkad.node.putNode
import java.net.SocketAddress
import java.nio.ByteBuffer

/**
 * Created by Christoph on 21.09.2016.
 */

enum class MessageType constructor(
    val ser: (msg: Message, buf: ByteBuffer) -> Unit,
    val deser: (buf: ByteBuffer) -> Message
) {

    PING({ msg, buf -> msg.writeTo(buf) }, {
        val rd = Message.readFrom(it)
        Ping(rd.seqId, rd.origin)
    }),
    PONG({ msg, buf -> msg.writeTo(buf) }, {
        val rd = Message.readFrom(it)
        Pong(rd.seqId, rd.origin)
    }),
    STORE({ msg, buf ->
        msg.writeTo(buf)
        (msg as Store)
        buf.putInt(msg.value.size)
        buf.put(msg.value)
    }, {
        val rd = Message.readFrom(it)
        val sz = it.int
        val bytes = ByteArray(sz).apply { it.get(this) }
        Store(rd.seqId, rd.origin, bytes)
    }),
    FIND_NODE({ msg, buf ->
        msg.writeTo(buf)
        buf.putIdentifier((msg as FindNode).target)
    }, {
        val rd = Message.readFrom(it)
        FindNode(rd.seqId, rd.origin, it.getIdentifier())
    }),
    FIND_VALUE({ msg, buf ->
        msg.writeTo(buf)
        buf.putIdentifier((msg as FindValue).key)
    }, {
        val rd = Message.readFrom(it)
        FindValue(rd.seqId, rd.origin, it.getIdentifier())
    }),
    NODE_REPLY({ msg, buf ->
        msg.writeTo(buf)
        msg as NodeReply
        buf.putShort(msg.nodes.size.toUShort().toShort())
        msg.nodes.forEach { buf.putNode(it) }
    }, {
        val rd = Message.readFrom(it)
        val numNodes = it.short.toUShort().toInt()
        val nodes = kotlin.collections.ArrayList<Node>(numNodes)
        for (i in 1..numNodes) {
            nodes.add(it.getNode())
        }
        NodeReply(rd.seqId, rd.origin, nodes)
    }),
    STORE_REPLY({ msg, buf -> msg.writeTo(buf) }, {
        val rd = Message.readFrom(it)
        StoreReply(rd.seqId, rd.origin)
    }),
    VALUE_REPLY({ msg, buf ->
        msg.writeTo(buf)
        (msg as ValueReply)
        buf.putInt(msg.value.size)
        buf.put(msg.value)
    }, {
        val rd = Message.readFrom(it)
        val sz = it.int
        val bytes = ByteArray(sz).apply { it.get(this) }
        ValueReply(rd.seqId, rd.origin, bytes)
    }),
    EXTENDED({ msg, buf -> msg.writeTo(buf) }, {
        val rd = Message.readFrom(it)
        val extendedType = it.short.toUShort()
        val deserializer = otherMap.get(extendedType)
        deserializer?.invoke(rd, it) ?: UnknownMessage(extendedType, rd.seqId, rd.origin)
    });

    companion object {
        val otherMap = mutableMapOf<UShort, (header: MessageHeader, buf: ByteBuffer) -> Message>()
    }
}

fun ByteBuffer.putMessage(msg: Message) {
    msg.type.ser.invoke(msg, this)
}

fun ByteBuffer.getMessage() = MessageType.values()[short.toInt()].deser(this)


data class MessageHeader(val seqId: Long, val origin: Node)
abstract class Message(val type: MessageType, val seqId: Long, val origin: Node) {
    override fun toString() = "${javaClass.name}(seqId=$seqId, origin=$origin)"

    open fun writeTo(buffer: ByteBuffer) {
        buffer.putShort(type.ordinal.toShort())
        buffer.putLong(seqId)
        buffer.putNode(origin)
    }

    companion object {
        fun readFrom(buf: ByteBuffer) = MessageHeader(buf.long, buf.getNode())
    }
}

class UnknownMessage(val extendedType: UShort, seqId: Long, origin: Node) : Message(MessageType.EXTENDED, seqId, origin)

abstract class ExtendedMessage(
    val extendedType: UShort,
    seqId: Long,
    origin: Node,
    deser: (header: MessageHeader, buf: ByteBuffer) -> Message
) :
    Message(MessageType.EXTENDED, seqId, origin) {
    init {
        MessageType.otherMap.putIfAbsent(extendedType, deser)
    }

    protected abstract fun serialize(buffer: ByteBuffer)
    override fun writeTo(buffer: ByteBuffer) {
        super.writeTo(buffer)
        buffer.putShort(extendedType.toShort())
        serialize(buffer)
    }
}

data class Packet(val msg: Message, val origin: SocketAddress)
