package at.tugraz.iaik.p2p.hkad.node

import at.tugraz.iaik.p2p.commons.CommonCrypto
import at.tugraz.iaik.p2p.commons.UDPChannel
import at.tugraz.iaik.p2p.hkad.ex.CodecException
import at.tugraz.iaik.p2p.hkad.ex.NotRunningException
import at.tugraz.iaik.p2p.hkad.ex.SeqIdMisMatch
import at.tugraz.iaik.p2p.hkad.id.Identifier
import at.tugraz.iaik.p2p.hkad.msg.*
import kotlinx.coroutines.CancellationException
import kotlinx.coroutines.delay
import kotlinx.coroutines.withTimeout
import kotlinx.coroutines.yield
import java.nio.ByteBuffer
import java.util.concurrent.LinkedBlockingQueue
import java.util.concurrent.TimeoutException
import java.util.concurrent.atomic.AtomicInteger

class Actor(private val node: LocalNode) {


    private val codec = Codec(node.config.privateKey)
    private val inTransit = AtomicInteger(0)
    private var running = true

    private inner class Buffers(val sendBuffer: ByteBuffer, val rxdBuffer: ByteBuffer) {
        suspend fun retire() {
            while (!bufferQ.offer(this)) yield()
        }
    }

    private val bufferQ = LinkedBlockingQueue<Buffers>(100).apply {
        while (remainingCapacity() > 0) add(
            Buffers(
                ByteBuffer.allocateDirect(node.config.datagramSize.toInt()),
                ByteBuffer.allocateDirect(node.config.datagramSize.toInt())
            )
        )
    }

    private suspend fun getBuffers(): Buffers {
        var buffers = bufferQ.poll()
        while (buffers == null) {
            yield()
            buffers = bufferQ.poll()
        }
        return buffers
    }


    suspend fun ping(receiver: Node) = send(Ping(CommonCrypto.RND.nextLong(), node), receiver) as Pong

    suspend fun findNode(target: Identifier, receiver: Node) =
        send(FindNode(CommonCrypto.RND.nextLong(), node, target), receiver) as NodeReply

    suspend fun findValue(key: Identifier, receiver: Node) =
        send(FindValue(CommonCrypto.RND.nextLong(), node, key), receiver)

    suspend fun store(value: ByteArray, receiver: Node) =
        send(Store(CommonCrypto.RND.nextLong(), node, value), receiver) as StoreReply


    suspend fun send(msg: Message, receiver: Node): Message {

        val buffers = getBuffers()

        val channel = UDPChannel(buffers.sendBuffer, buffers.rxdBuffer)
        try {
            inTransit.addAndGet(1)
            val resp = doSend(msg, channel, receiver).msg
            node.routingTable.add(NodeInfo(resp.origin, System.currentTimeMillis().toULong()))
            return resp
        } finally {
            inTransit.addAndGet(-1)
            channel.close()
            buffers.retire()
        }
    }


    @Throws(TimeoutException::class, NotRunningException::class)
    private suspend fun doSend(msg: Message, channel: UDPChannel, dest: Node): Packet {

        if (!running) throw NotRunningException()

        repeat(node.config.retries.toInt()) { iteration ->
            if (running) {
                try {
                    node.logger.trace { "sending ${msg.type} to ${dest.id} (${dest.addr})" }

                    codec.encode(msg, channel.sendBuffer)


                    channel.transmit(dest.addr)

                    try {
                        //TODO check seqID
                        return withTimeout(node.config.networkTimeoutMs.toLong()) {
                            val rxd = channel.receive()
                            val message = codec.decode(rxd.data)
                            if (message.seqId != msg.seqId)
                                throw SeqIdMisMatch(message.seqId, msg.seqId)
                            return@withTimeout Packet(message, rxd.src)
                        }

                    } catch (to: CancellationException) {
                        node.logger.warn { "sending ${msg.type} to ${dest.id} timed out (try ${iteration + 1})" }
                    } catch (mm: SeqIdMisMatch) {
                        node.logger.warn("response to ${msg.type} does not match seqID: actual: ${mm.actual}, expected: ${mm.expected}")
                    }

                } catch (e: CodecException) {
                    throw e
                }

                delay(10)
            }
        }
        if (running) throw TimeoutException()
        throw NotRunningException()
    }

}