package at.tugraz.iaik.p2p.hkad.node

import at.tugraz.iaik.p2p.hkad.config.Config
import at.tugraz.iaik.p2p.hkad.ex.HashMismatch
import at.tugraz.iaik.p2p.hkad.id.Identifier
import at.tugraz.iaik.p2p.hkad.id.getIdentifier
import at.tugraz.iaik.p2p.hkad.id.putIdentifier
import at.tugraz.iaik.p2p.hkad.msg.ExtendedMessage
import at.tugraz.iaik.p2p.hkad.msg.Message
import at.tugraz.iaik.p2p.hkad.msg.NodeReply
import at.tugraz.iaik.p2p.hkad.msg.ValueReply
import com.google.common.primitives.UnsignedBytes
import kotlinx.coroutines.Deferred
import kotlinx.coroutines.GlobalScope
import kotlinx.coroutines.async
import kotlinx.coroutines.awaitAll
import mu.NamedKLogging
import java.net.InetSocketAddress
import java.nio.ByteBuffer
import java.util.*
import kotlin.coroutines.CoroutineContext
import kotlin.coroutines.EmptyCoroutineContext


fun ByteBuffer.putNode(node: Node) {
    putIdentifier(node.id)

    val addr = node.addr.hostName.toCharArray()
    val numChars = addr.size.toShort()

    putShort(numChars)

    addr.forEach { putChar(it) }

    putShort(node.addr.port.toUShort().toShort())
    putShort(node.nonceRange.toShort())
}

fun ByteBuffer.getNode(): Node {
    val id = getIdentifier()

    val numChars = getShort().toUShort().toInt()

    val addrChars = CharArray(numChars) { getChar() }

    val port = getShort().toUShort().toInt()
    val nonceRange = getShort().toUShort()
    return Node(id, InetSocketAddress(String(addrChars), port), nonceRange)
}

data class Value(val id: Identifier, val data: ByteArray?, val origin: Node?, val closestNodes: SortedSet<Node>) {
    override fun equals(other: Any?): Boolean {
        if (this === other) return true
        if (javaClass != other?.javaClass) return false

        other as Value

        if (id != other.id) return false
        if (data != null) {
            if (other.data == null) return false
            if (!data.contentEquals(other.data)) return false
        } else if (other.data != null) return false
        if (origin != other.origin) return false

        return true
    }

    override fun hashCode(): Int {
        var result = id.hashCode()
        result = 31 * result + (data?.contentHashCode() ?: 0)
        result = 31 * result + (origin?.hashCode() ?: 0)
        return result
    }

}

open class Node(val id: Identifier, val addr: InetSocketAddress, val nonceRange: UShort) {

    val virtualIDs by lazy {
        val buf = ByteBuffer.allocate(id.keyBytes.size + 4)
        buf.put(id.keyBytes)
        Array<Identifier>(nonceRange.toInt()) {
            id.fromData(buf.putInt(id.keyBytes.size, it))
        }

    }


    internal val logger = NamedKLogging("$id").logger

    override fun equals(other: Any?): Boolean {
        if (other !is Node)
            return false
        return id == other.id
    }

    override fun toString(): String {
        return "${javaClass.simpleName} $id"
    }
}


open class LocalNode(
    val config: Config,
    private val actorCtx: CoroutineContext = EmptyCoroutineContext,
    reactorCtx: CoroutineContext = EmptyCoroutineContext,
    incomingMsgListener: (suspend (
        msg:
        ExtendedMessage
    ) -> ExtendedMessage)? = null
) : Node(config.id, config.listenAddr, config.nonceRange) {


    internal val routingTable = RoutingTable(this)
    val actor = Actor(this)
    private val reactor = Reactor(this, reactorCtx, incomingMsgListener)
    internal val store = InMemoryStorage()


    data class PutInfo(val key: Identifier, val location: Identifier)

    suspend fun put(data: ByteArray): PutInfo {
        val id = id.fromData(ByteBuffer.wrap(data))
        val find = find(id, false)
        if (find.origin != null)
            return PutInfo(id, find.origin.id)
        actor.store(data, find.closestNodes.first())
        return PutInfo(id, find.closestNodes.first().id)
    }

    suspend fun get(target: Identifier) = find(target, false)

    internal suspend fun find(target: Identifier, lookingForNode: Boolean): Value {
        logger.debug { "getting closest nodes for $target" }
        logger.debug { routingTable.toString() }
        val cmp = UnsignedBytes.lexicographicalComparator()
        val comparator = Comparator<Node> { o1, o2 ->
            cmp.compare(
                (o1.id xor target),
                (o2.id xor target)
            )
        }
        val toAsk = routingTable.closestFor(target, !lookingForNode)
        val checked: SortedSet<Node> = sortedSetOf(comparator, this)
        val result = findClosest(target, toAsk.map { it.node }.toSortedSet(comparator), checked, lookingForNode)
        return Value(id = target, data = result?.value, origin = result?.origin, closestNodes = checked)
    }

    protected tailrec suspend fun findClosest(
        target: Identifier,
        toAsk: SortedSet<Node>,
        checked: SortedSet<Node>, lookingForNode: Boolean
    ): ValueReply? {
        if (lookingForNode) try {
            toAsk.first { id == target }?.let { checked.add(it) }
        } catch (e: NoSuchElementException) {
            //not contained
        }
        if (lookingForNode) if (target in checked.map { it.id })
            return null
        if (toAsk.isEmpty()) return null

        val jobs = mutableListOf<Deferred<Message>>()

        val currentlyAsking = toAsk.take(config.alpha.toInt())
        toAsk.removeAll(currentlyAsking)
        currentlyAsking.forEach {
            logger.debug { "asking ${it.id} for $target" }
            jobs.add(GlobalScope.async(actorCtx) {
                if (lookingForNode) actor.findNode(target, it)
                else actor.findValue(target, it)
            })
        }
        jobs.awaitAll().forEach {
            logger.trace { "\t adding ${it.origin.id} ${it.origin.addr} to checked" }
            checked.add(it.origin)
            if (it is NodeReply) {
                logger.trace { "\t adding ${it.nodes.map { "${it.id} ${it.addr}" }} to toAsk" }
                toAsk.addAll(it.nodes)
            } else if (it is ValueReply) {
                val calculatedHash = id.fromData(ByteBuffer.wrap(it.value))
                if (calculatedHash == target) {
                    logger.debug { "found value for $target at ${it.origin}" }
                    return it
                } else
                    throw HashMismatch(calculatedHash, target)

            }
        }
        logger.trace { "toAsk contains: $toAsk\nremoving $checked from toAsk" }
        toAsk.removeAll(checked)
        logger.trace { "learned ${toAsk.map { it.id }} ${toAsk.map { it.addr }}" }
        return findClosest(target, toAsk, checked, lookingForNode)

    }


    suspend fun bootstrap(node: Node) {
        val bootstrapNode = actor.ping(node).origin

        val nodes = actor.findNode(id, bootstrapNode).nodes
        nodes.forEach { actor.ping(it) }

        routingTable.refresh()

    }

    suspend fun close() {
        reactor.close()
    }


    inner class InMemoryStorage : Storage() {
        private val map = HashMap<Identifier, ByteArray>()
        override fun store(data: ByteArray): Identifier {
            val id = this@LocalNode.id.fromData(ByteBuffer.wrap(data))
            map[id] = data
            return id
        }

        override fun get(id: Identifier): ByteArray? = map[id]
        override fun contains(id: Identifier): Boolean = map.containsKey(id)

    }

}


