package at.tugraz.iaik.p2p.hkad.node

import at.tugraz.iaik.p2p.commons.UDPChannel
import at.tugraz.iaik.p2p.hkad.id.Identifier
import at.tugraz.iaik.p2p.hkad.msg.*
import kotlinx.coroutines.GlobalScope
import kotlinx.coroutines.cancelAndJoin
import kotlinx.coroutines.isActive
import kotlinx.coroutines.launch
import java.net.InetSocketAddress
import java.net.SocketAddress
import java.nio.ByteBuffer
import kotlin.coroutines.CoroutineContext

class Reactor(
    private val node: LocalNode,
    ctx: CoroutineContext,
    private val incomingMsgListener: (suspend (msg: ExtendedMessage) -> ExtendedMessage)? = null
) {

    private val codec = Codec(node.config.privateKey)
    private val channel = UDPChannel(ByteBuffer.allocateDirect(node.config.datagramSize.toInt())).apply {
        bind(node.config.listenAddr)
    }

    private val server = GlobalScope.launch(ctx) {
        while (isActive) {
            try {
                val rxd = channel.receive()
                handle(rxd.src, rxd.data)
            } catch (ex: Throwable) {
                if (!channel.isClosed) throw ex
            }
        }
    }

    private suspend fun handle(address: InetSocketAddress, buffer: ByteBuffer) {
        val msg = codec.decode(buffer)
        node.logger.trace { "got ${msg.type} message from ${msg.origin}" }
        node.routingTable.add(NodeInfo(msg.origin, System.currentTimeMillis().toULong()))

        when {
            msg.type === MessageType.PING -> handlePing(msg, address)
            msg.type === MessageType.FIND_NODE -> handleFindNode(msg as FindNode, address)
            msg.type === MessageType.FIND_VALUE -> handleFindValue(msg as FindValue, address)
            msg.type === MessageType.STORE -> handleStore(msg as Store, address)
            msg.type=== MessageType.EXTENDED -> {
                val resp = incomingMsgListener?.invoke(msg as ExtendedMessage)
                if (resp != null)
                    send(resp, msg.origin.id, address)
            }
        }
    }


    private suspend fun send(msg: Message, receiverId: Identifier, address: SocketAddress) {
        node.logger.trace { "responding with ${msg.type} to $receiverId ($address)" }
        codec.encode(msg, channel.sendBuffer)
        channel.transmit(address)
    }

    private suspend fun handleStore(msg: Store, address: SocketAddress) {
        node.store.store(msg.value)
        send(StoreReply(msg.seqId, node), msg.origin.id, address)
    }

    private suspend fun handleFindValue(msg: FindValue, address: SocketAddress) {
        if (node.store.contains(msg.key)) {
            send(ValueReply(msg.seqId, node, node.store[msg.key]!!), msg.origin.id, address)
            return
        }

        val closest = node.routingTable.closestFor(msg.key, true).map { it.node }
        send(NodeReply(msg.seqId, node, closest), msg.origin.id, address)
    }

    private suspend fun handleFindNode(msg: FindNode, address: SocketAddress) {
        val closest = node.routingTable.closestFor(msg.target, false).map { it.node }
        send(NodeReply(msg.seqId, node, closest), msg.origin.id, address)
    }

    private suspend fun handlePing(msg: Message, address: SocketAddress) {
        send(Pong(msg.seqId, node), msg.origin.id, address)
    }

    suspend fun close() {
        channel.close()
        server.cancelAndJoin()
    }

}