package at.tugraz.iaik.p2p.hkad.node

import at.tugraz.iaik.p2p.hkad.id.Identifier
import com.google.common.primitives.UnsignedBytes
import java.util.*
import java.util.concurrent.ConcurrentSkipListSet
import java.util.concurrent.TimeoutException


data class NodeInfo(val node: Node, var lastSeen: ULong, val virtID: UShort? = null) : Comparable<NodeInfo> {
    override fun compareTo(other: NodeInfo): Int {
        if (node == other.node)
            return 0
        return lastSeen.compareTo(other.lastSeen)
    }

    override fun equals(other: Any?): Boolean {
        if (other !is NodeInfo) return false
        return node == other.node
    }
}

class RoutingTable(private val node: LocalNode) {


    private val buckets = Array(node.config.id.numBits.toInt()) { Bucket(it) }
 //   private val virtBuckets = Array(node.config.id.numBits.toInt()) { Bucket(it) }

    fun closestFor(target: Identifier, virt: Boolean): SortedSet<NodeInfo> {
        val cmp = UnsignedBytes.lexicographicalComparator()
        val comparator = Comparator<NodeInfo> { o1, o2 ->
            cmp.compare(
                (o1.node.id xor target),
                (o2.node.id xor target)
            )
        }
        return /*if (!virt) {*/            buckets.flatMap { it.nodes }.sortedWith(comparator).take(node.config.k.toInt()).toSortedSet(comparator)
      //  } else {
        //    val virtComparator = kotlin.Comparator<NodeInfo> { o1, o2 ->
          //      cmp.compare(
            //        if (o1.virtID == null) o1.node.id xor target else o1.node.virtualIDs[o1.virtID.toInt()] xor target,
              //      if (o2.virtID == null) o2.node.id xor target else o2.node.virtualIDs[o2.virtID.toInt()] xor target
                //)
           // }
           // virtBuckets.flatMap { it.nodes }.sortedWith(virtComparator).distinctBy { it.node }
           //     .take(node.config.k.toInt()).toSortedSet(virtComparator)
       // }
    }

    suspend fun add(info: NodeInfo) {
        if (info.node.id == this.node.id)
            return
        val bucketId = this.node.config.id.distanceTo(info.node.id)
        val bucket = buckets[bucketId]
        bucket.add(info)

      //  if (node.nonceRange == 0u.toUShort()) {
      //      val virtBucketId = this.node.config.id.distanceTo(info.node.id)
      //      val virtBucket = virtBuckets[virtBucketId]
      //      virtBucket.add(NodeInfo(info.node, info.lastSeen, null))
      //  } else {
      //      for (i in 0..info.node.nonceRange.toInt()) {
      //          val virtBucketId = this.node.config.id.distanceTo(info.node.virtualIDs[i])
      //          val virtBucket = virtBuckets[virtBucketId]
      //          virtBucket.add(NodeInfo(info.node, info.lastSeen, i.toUShort()))
      //      }
      //  }
    }

    suspend fun refresh(): Boolean {
        val bucket = try {
            buckets.first { it.nodes.isNotEmpty() }
        } catch (e: NoSuchElementException) {
            return false
        }
        val minDist = bucket.index + 1
        for (i in minDist until buckets.size) {
            val lookupId = node.config.id.generateBydistance(i)
            try {
                node.find(lookupId, lookingForNode = true)
            } catch (t: Throwable) {
                node.logger.warn { t.localizedMessage }
            }
        }
        return true
    }

    inner class Bucket(val index: Int) {
        val nodes = ConcurrentSkipListSet<NodeInfo>()
        val replacementNodes = ConcurrentSkipListSet<NodeInfo>()

        suspend fun add(info: NodeInfo) {
            if (info in nodes) {
                nodes.add(info)
                this@RoutingTable.node.logger.trace { "Bucket $index Refreshing $info (already known)" }
                return
            }
            if (nodes.size < this@RoutingTable.node.config.k.toInt()) {
                nodes.add(info)
                this@RoutingTable.node.logger.trace { "Bucket $index Adding $info" }
                return
            }
            val last = nodes.last()
            if (System.currentTimeMillis().toULong() - last.lastSeen < 15000uL) {
                this@RoutingTable.node.logger.trace { "Bucket $index Not pinging ${last.node.id}" }
                return
            }
            try {
                this@RoutingTable.node.logger.trace { "Bucket $index Pinging ${last.node.id}" }
                this@RoutingTable.node.actor.ping(last.node)
                last.lastSeen = System.currentTimeMillis().toULong()
                nodes.add(last)
                replacementNodes.add(info)
                if (replacementNodes.size > this@RoutingTable.node.config.k.toInt()) {
                    replacementNodes.remove(replacementNodes.last())
                }
                return

            } catch (e: TimeoutException) {
                this@RoutingTable.node.logger.trace { "Bucket $index Removing unresponsive ${last.node.id}; adding ${info.node.id}" }
                nodes.remove(last)
                nodes.add(info)
                return
            }
        }

        override fun toString(): String {
            return "Bucket $index  ${nodes.size} node(s): ${nodes.map {
                "${it.node.id} ${it.node.addr} (${node.id.distanceTo(
                    it.node.id
                )})"
            }}"
        }
    }

    override fun toString() = "${node.config.id} RoutingTable [\n ${
    buckets.filter { it.nodes.isNotEmpty() }.map {
        "$it)\n"
    }}"
}